#------------------------------------------------------------------------------
# snackAmp - Catalog package
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#------------------------------------------------------------------------------
set [file rootname [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/catalog.tcl,v 1.87 2005/06/05 01:56:51 wilkason Exp $
}
namespace eval catalog {
   variable lastList [list]
   variable indexing 0
   variable flag 0
   variable indexUnits
   variable fcount 0
   array set indexUnits {
      Minutes 60
      Hours 3600
      Days 86400
      Weeks 604800
   }
   variable S
   upvar ::snackAmpSettings S
}
#------------------------------------------------------------------------------
# Function   : catalog::saveAs
# Description: Save the current file as some format
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::saveAs {args} {
   set file [snack::getSaveFile]
   if {[string length $file]==0} {
      return
   }
   # Stop (should work even if playing)
   # Clone current sound
   set snd [snack::sound -file [soundControl::fileName]]
   # Get extension
   set ext [string tolower [string range [file extension $file] 1 end]]
   ##
   # Status Window
   #
   # Save the file
   if {[catch {
      switch -- $ext {
         "ogg" {
            #W: Prompt for -nominalbitrate, -maxbitrate, -minbitrate or use existing
            switch -- [$snd config -fileformat] {
               MP3 {set rate [$snd config -bitrate]}
               OGG {set rate [$snd config -nominalbitrate]}
               WAV {set rate [expr {[$snd config -rate]}]}
               default {set rate 0}
            }
            if {$rate>0} {
               waitWindow::show "Saving $file ... 0% complete"
               $snd write $file -fileformat $ext -nominalbitrate $rate -progress catalog::waiter
            } else {
               waitWindow::show "Saving ... 0% complete"
               $snd write $file -fileformat $ext -progress catalog::waiter
            }
         }
         default {
            $snd write $file -fileformat $ext -progress catalog::waiter
         }
      }
   } result]} {
      tk_messageBox -icon warning -type ok -title "Save Error" -message $result
   }
   # Make sure it is in the master list
   db::cacheFileName $file
   $snd destroy
   waitWindow::close
}
#------------------------------------------------------------------------------
# Function   : catalog::waiter
# Description: Update global used for wait box
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::waiter {cmd percent} {
   set percent [expr {round($percent*1000)/10.0}]
   waitWindow::show  "Saving ... $percent% complete"
}

#------------------------------------------------------------------------------
# Return true if this files has an extension that we car about
#------------------------------------------------------------------------------
proc catalog::isTrackFile {file} {
   global taExtensions
   return [info exists taExtensions([file extension $file])]
}
#------------------------------------------------------------------------------
# Function   : catalog::MonitorScanFolders
# Description: Monitor the base folders for incoming data, useful
#              for napaster adding files.
# Author     : Tom Wilkason
# Date       : 12/30/2000
#------------------------------------------------------------------------------
proc catalog::MonitorScanFolders {args} {
   global snackAmpSettings TA
   if {$snackAmpSettings(MonitorScanFolders)} {
      # TODO: update folders in APL also
      foreach {Folder} $snackAmpSettings(IncomingDirs) {
         foreach {file} [eval glob -nocomplain -directory {$Folder} [globAllPattern $snackAmpSettings(ScanFolderDepth)]] {
            if {[catalog::isTrackFile $file]} {
               if {[db::whereInDb $file]==-1} {
                  catalog::addNewEntry $file
               }
               Update
            }
         }
      }
   }
   after cancel catalog::MonitorScanFolders
   after 60000 catalog::MonitorScanFolders
}

#------------------------------------------------------------------------------
# Function   : catalog::fixMediaPath
# Description: Make sure the directories in the media path don't have trailing slashes
#              needed by tree matching functions
# Author     : Tom Wilkason
# Date       : 1/11/2001
# TBD: we can revalidate the path here
#------------------------------------------------------------------------------
proc catalog::fixMediaPath {args} {
   cleanPathList ::snackAmpSettings(IncomingDirs) 1
   cleanPathList ::snackAmpSettings(mediaDirs) 0 ;#1
   cleanPathList ::snackAmpSettings(mp3Player) 0 ;#1
}
#----------------------------------------------------------------------------
# Rescan all the media dirs after asking for permission.
#----------------------------------------------------------------------------
proc catalog::IndexAskFirst {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to index the following folders?\n$::snackAmpSettings(mediaDirs)"]=="yes"} {
      catalog::Index
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::Index
# Description: Reread the directories for files, include subdirectories
#              For each folder, add a menu entry
# Author     : Tom Wilkason
# Date       : 12/26/2000
#------------------------------------------------------------------------------
proc catalog::Index {{Folder ""}} {
   global TA snackAmpSettings
   variable fcount 0
   Trace
   Wait 1
   set here [pwd]
   # If a folder is specifed, catalog all media dirs, otherwise just
   # catalog the specified folder retaining information on other folders.
   if {$Folder == ""} {
      waitWindow::show "Indexing tracks\nPlease wait...\n" 1
      set saveNames [aplFolder::savePattern *]  ;# want folder also
      aplFolder::stripBranch *
      # Get a list of all the files intoList
      # Only scan the known mediaDirs
      foreach dir $TA(mediaDirs) {
         # Flag all as unmounted till recached
         db::unmountBranch [file join $dir *]
         indexRecurse $dir
         db::removeUnmountedFiles [file join $dir *]
         if {[waitWindow::ifbreak]} {
            break
         }
      }
      waitWindow::show "Verifying/Reading Catalogs"
      ##
      # Update the APL settings
      # (should really restore the old memory settings)
      # Restore prior settings, note it indexing
      # is done prior to first time APL use, then
      # all apl values are 0. Need to fiX
      aplFolder::restoreNames $saveNames
      # Rebuild everything since we indexed everything
      waitWindow::show "Cleaning Up ... Almost Done"
      if {[APLExists]} {
         APL::reBuildTree
      }
   } else {
      # Set master list to all but those files in this branch
      waitWindow::show "Indexing tracks on\n$Folder\n"
      # Flag these files as unmounted
      # Then clear folders
      set saveNames [aplFolder::savePattern $Folder*]  ;# want folder also
      aplFolder::stripBranch [file join $Folder *]
      waitWindow::bump
      # Flag tracks in db as unmounted
      db::unmountBranch [file join $Folder *]
      # Rescan selected folder (restore any apl folder settings)
      waitWindow::bump
      indexRecurse $Folder
      # Any tracks not flagged as mounted here no longer exist, remove them
      db::removeUnmountedFiles [file join $Folder *]
      waitWindow::bump
      # Restore prior settings to folders that exist
      aplFolder::restoreNames $saveNames

      #BUG: stale (i.e. no longer existing) teFolder settings can still hang around

      ## Add (or ignore) new items for each entry
      if {[APLExists]} {
         # remove branch
         APL::delItem $Folder
         # add it and the new children back
         if {[file exists $Folder]} {
            APL::newItem $Folder
            foreach folder [aplFolder::foldersMatchingPattern [file join $Folder *] 1] {
               APL::newItem $folder
            }
         }
      }
   }
   ##
   # Build the folder list for the menu
   #
   waitWindow::close
   Wait 0
   # Save the files
   after 100 db::Commit
}
#------------------------------------------------------------------------------
# Escape characters in file/folder names that cause problems
# with globbing and regexp.
#------------------------------------------------------------------------------
proc catalog::globEscape {dir} {
   return [string map {( \\( ) \\) \{ \\{ \} \\} \] \\] \[ \\[} $dir]
}
#------------------------------------------------------------------------------
# Recursively scan a directory and build a list of all files
#------------------------------------------------------------------------------
proc catalog::indexRecurse {dir} {
   variable fcount
   global TA LB snackAmpSettings
   set dirList [list]
   # Show banner every once in a while
   set hasSome 0
   set alwaysCatalog $snackAmpSettings(alwaysCatalog)
   if {$alwaysCatalog} {
      waitWindow::show "Indexing/Cataloging tracks on\n$dir\n[commify $fcount] Indexed..." 1
   } else {
      waitWindow::show "Indexing tracks on\n$dir\n[commify $fcount] Indexed..." 1
   }
   # Allow {} and [] in names (glob uses them as special chars)
   set gdir [globEscape [file join $dir *]]
   if {$snackAmpSettings(fastscan)} {
   # This is the fastest recursive globbing method I can come up with
      foreach file [glob -nocomplain $gdir] {
         # This is problematic if folder ending in .mp3
         if {[catalog::isTrackFile $file]} {
            # Cache if needed
            if {$alwaysCatalog} {
               unset -nocomplain Data
               db::cacheFile $file Data 2
               incr hasSome
            } else {
               db::cacheFileName $file 1
               incr hasSome
            }
         } elseif {[file isdirectory $file]} {
            lappend dirList $file
            incr hasSome
         }
      }
   } else {
      # This is a bit slower due to additional directory checks
      # but will see ALL directories
      foreach file [glob -nocomplain $gdir] {
         if {[file isdirectory $file]} {
            lappend dirList $file
            incr hasSome
         } elseif {[catalog::isTrackFile $file]} {
            db::cacheFileName $file 1
            incr hasSome
         }
      }
   }
   # Only set the folder (as existing) for the tree if
   # it was non-empty
   if {$hasSome} {
      aplFolder::Set $dir 0      ;# Add this dir, if new will be turned of
      incr fcount $hasSome
   }
   # process each directory recursively
   foreach {folder} $dirList {
      indexRecurse $folder
   }
}
#----------------------------------------------------------------------------
# Background index a folder
#----------------------------------------------------------------------------
proc catalog::bgIndex {args} {
   global TA
   variable indexing
   variable indexUnits
   variable S
   if {$indexing} {
      saLog "Background index in progress"
   } else {
      saLog "Running a background index"
      set indexing 1
       foreach dir $TA(mediaDirs) {
          if {[catalog::isFixedFolder $dir]} {
             set S(indexing) "Cleaning $dir"
             aplFolder::removeUnusedNodes [file join $dir *]
             bgIndexRecurse $dir
             db::removeUnmountedFiles [file join $dir *]
             # TODO: need to add any apl nodes that are new here
          }
       }
      set indexing 0
      set S(indexing) "Idle"
      set S(lastIndex)              [clock seconds]
      set S(lastIndexFormatted)     [clock format [clock seconds]]
      set S(nextIndex)              [expr {[clock seconds]+$S(indexHours)*$indexUnits($S(indexUnits))}]
      set S(nextIndexFormatted)     [clock format $S(nextIndex)]
   }
}

#------------------------------------------------------------------------------
# Routinely run to see if a background index should be run
# TODO: Need to remove stale folders from tefolders
#------------------------------------------------------------------------------
proc catalog::bgIndexCheck {args} {
   variable indexing
   variable S
   after cancel catalog::bgIndexCheck
   set now [clock seconds]
   if {[string equal $S(indexStart) "0"]} {
      set secStart $now
   } elseif {[catch {clock scan $S(indexStart)} secStart] } then {
      saLog "Invalid time format: $snackAmpSettings(indexStart) for Index Start Time"
      set secStart $now
   }
   if {[string equal $S(indexStop) "0"]} {
      set secStop $now
   } elseif {[catch {clock scan $S(indexStop)} secStop] } then {
      saLog "Invalid time format: $snackAmpSettings(indexStop) for Index Stop Time"
      set secStop $now
   }
   if {$S(indexBackground) && $S(indexHours)>0 && $indexing==0} {
      # valid to check to see if we can run
      if {$now >= $S(nextIndex) && ($now >= $secStart) && ($now <= $secStop)} {
         after 0 catalog::bgIndex
      }
   }
   # Check again in five minutes
   after 300000 catalog::bgIndexCheck
}
#----------------------------------------------------------------------------
# Background index function
#----------------------------------------------------------------------------
proc catalog::bgIndexRecurse {dir} {
   global TA snackAmpSettings
   variable flag
   if {$snackAmpSettings(indexBackground)==0} {
      update
      return
   }
   set bgCatalog $snackAmpSettings(bgCatalog)
   set dirList [list]
   set hasSome 0
   # Allow {} and [] in names (glob uses them as special chars)
   set gdir [globEscape [file join $dir *]]
   # Flag folder as unmounted, recaching will remount files
   set snackAmpSettings(indexing) $dir
   db::mountFolder $dir 0
   # Use effecient method
   foreach file [glob -nocomplain $gdir] {
      # This is problematic if folder ending in .mp3
      update
      if {[catalog::isTrackFile $file]} {
         # Cache if needed, use timestamps on last catalog vs
         # the time on the file and the full data tag.
         if {$bgCatalog} {
            unset -nocomplain Data
            db::cacheFile $file Data 2
            incr hasSome
         } else {
            if {[db::cacheFileName $file 1] == -1} {
               saLog "Cannot Catalog $file"
            } else {
               incr hasSome
            }
         }
      } elseif {[file isdirectory $file]} {
         lappend dirList $file
         incr hasSome
      }
   }
   # Only set the folder (as existing) for the tree if
   # it was non-empty
   if {$hasSome} {
      aplFolder::SetIfBlank $dir 0      ;# Add this dir, if new will be turned of
   }
   # process each directory recursively
   foreach {folder} $dirList {
      # delay a bit in the background
      after 100 set catalog::flag 1
      vwait catalog::flag
      bgIndexRecurse $folder
   }
}
#----------------------------------------------------------------------------
# Verify all the mediadirs and set the mounted flag appropriatly
#----------------------------------------------------------------------------
proc catalog::validateMediaPaths {args} {
   Trace
   global TA snackAmpSettings
   set TA(mediaDirs) [list]
   set TA(mediaDirPair) [list]
   ##
   # Check hard folders
   foreach {Path} $snackAmpSettings(mediaDirs) {
      if {[file isdirectory $Path]} {
         lappend TA(mediaDirs) $Path
         lappend TA(mediaDirPair) $Path "Fixed Folders" 0
         db::flagMountedFolders $Path 1
      } else {
         saLog "Media folder '$Path' not found (is it mounted?)"
         db::flagMountedFolders $Path 0
      }
   }
   ##
   # Check attached media
   foreach {Path} $snackAmpSettings(mp3Player) {
      if {[file isdirectory $Path]} {
         lappend TA(mediaDirs) $Path
         lappend TA(mediaDirPair) $Path "Attached Folders" 2
         db::flagMountedFolders $Path 1
      } else {
         saLog "Media '$Path' is not attached"
         db::flagMountedFolders $Path 0
      }
   }
   return "" ;# always succeed
}
#------------------------------------------------------------------------------
# Function   : catalog::manageMediaFolders
# Description: Bring up dialog then update folder caches if needed
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::manageMediaFolders {args} {
   package require saSettings
   global snackAmpSettings
   set oldMedia $::TA(mediaDirs)
   folderManager::launch mediaDirs
   # Blocked here until GUI closes
   catalog::fixMediaPath
   catalog::validateMediaPaths
   if {[APLExists] && ![string equal $oldMedia $::TA(mediaDirs)]} {
      APL::reBuildTree
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::managePlayerFolders
# Description: Bring up dialog then update mp3 player caches if needed
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::managePlayerFolders {args} {
   package require saSettings
   global snackAmpSettings
   set oldMedia $::TA(mediaDirs)
   folderManager::launch mp3Player
   # Blocked here until GUI closes
   catalog::fixMediaPath
   catalog::validateMediaPaths
   if {[APLExists] && ![string equal $oldMedia $::TA(mediaDirs)]} {
      APL::reBuildTree
   }
}
#------------------------------------------------------------------------------
# Function   : playlistStartup
# Description: Populate the playlist with the appropriate files
# Author     : Tom Wilkason
# Date       : 6/5/2002
#------------------------------------------------------------------------------
proc catalog::playlistStartup {args} {
   Trace
   global snackAmpSettings TA
   ##
   # Rescan if set true or the cache read failed
   #
   if {$snackAmpSettings(AutoRescan) || $TA(rescan)} {
      catalog::Index
   }
   ;# Validate paths
   catalog::validateMediaPaths
   # Read cache files, now restore playlist if required
   # otherwise just toss up music based on default AutoPlaylist file
   if {$snackAmpSettings(restorePlayList)} {
      waitWindow::show "Restoring Playlist"
      if {[catch {catalog::ReadPlaylistCache} result]} {
         waitWindow::show "Reading [file tail $snackAmpSettings(lastFolderFile)]"
         readAPLFile $snackAmpSettings(lastFolderFile) 1
         saLog "No play cache file found\n$result"
      }
   } else {
      waitWindow::show "Reading [file tail $snackAmpSettings(lastFolderFile)]"
      readAPLFile $snackAmpSettings(lastFolderFile) 1
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::RenameFile
# Description: Try to rename/move a file, handle sound objects if they are playing
#              resume play if just renamed a playing file
# Author     : Tom Wilkason
# Date       : 3/22/2002
#------------------------------------------------------------------------------
proc catalog::RenameFile {oldName newName} {
   Trace
   set wasPlaying 0
   ##
   # Make sure move is enabled
   #
   if {$::snackAmpSettings(dontMoveFiles)} {
      return -code error "\nYou don't have permission to rename or move $oldName"
   }
   ##
   # If the current song playing is the one to be renamed
   # we must unlink it from snack then later relink it
   set justCaseChange [expr {[string equal -nocase $oldName $newName] && ($::tcl_platform(platform) == "windows")}]
   if {[file exists $newName] && !$justCaseChange} {
      if {[file isdirectory $newName]} {
         set res [tk_messageBox -icon question -type yesno -title "Target Exists" \
            -message "Warning: Folder $newName exists\nis it OK to overwrite any duplicate files?"]
      } else {
         catalog::trackFullData $oldName oldInfo
         catalog::trackFullData $newName newInfo
         # oldName is really latest file, newName will be existing file
         set textInfo "
     \tNew ->\tExisting
Len: \t$oldInfo(duration)\t$newInfo(duration)
Rate:\t$oldInfo(Rate)\t$newInfo(Rate)
Size:\t$oldInfo(size)\t$newInfo(size)
"
         if {($oldInfo(size) == $newInfo(size))} {
            set res [tk_messageBox -icon question -type yesno -title "File Exists" \
               -message "Warning: $newName exists\nand appears to be identical, do you still want to overwrite it?"]
         } else {
            set res [tk_messageBox -icon question -type yesno -title "File Exists" \
               -message "Warning: $newName\nexists, do you want to overwrite it?\n$textInfo"]
         }
      }
      if {$res != "yes"} {
         return -code error "$newName File Exists"
      }
   }
   set wasPlaying [soundControl::isPlayingTrack $oldName location]
   ##
   # Rename the file, if on windows then allow of a case insensitive name
   # Note: -forces allows overwritting of files but supresses errors if
   # the overwrite failed. This messes up upper error handling.
   if {$justCaseChange} {
      if {[catch {file rename $oldName $oldName.tmp} result]} {
         return -code error $result
      } else {
         if {[catch {file rename $oldName.tmp $newName} result]} {
            return -code error $result
         }
      }
   } else {
      # User may have inserted path separators to move into new folder
      set dirPart [file dirname $newName]
      if {![file exists $dirPart]} {
         if {[catch {file mkdir $dirPart} result]} {
            return -code error $result
         }
      }
      if {[file isdirectory $oldName] && [file isdirectory $newName]} {
#         saLog "Moving folders '$oldName' to existing folder '$newName'"
         set srcs [glob -nocomplain -directory $oldName *]
         set errors 0
         set Result ""
         foreach {file} $srcs {
            if {[catch {file rename -force -- $file $newName} result]} {
               append Result "\n$result"
               incr errors
            }
         }
         if {$errors > 0} {
            return -code error " ($errors) Move Errors $Result"
         }
         # Old folder will be left, can remove it if empty (if all went well)
         if {[file exists $oldName] && [llength [glob -nocomplain -directory $oldName *]]==0} {
            catch {file delete $oldName}
         }
       } else {
         # The target folder is new
         if {[catch {file rename -force -- $oldName $newName} result]} {
            return -code error $result
         }
      }
   }
   ##
   # Resume playing if active song was renamed
   #
   if {$wasPlaying} {
      soundControl::resumePlayingTrack $newName $location
   }
}
#----------------------------------------------------------------------------
# Copy a file, typically to a media player
#----------------------------------------------------------------------------
proc catalog::copyFile {oldName newName} {
   Trace
   ## Check for overwrites
   if {[file exists $newName]} {
      if {[file isdirectory $newName]} {
         set res [tk_messageBox -icon question -type yesno -title "Target Exists" \
            -message "Warning: Folder $newName exists\nis it OK to overwrite any older duplicate files?"]
      } else {
         catalog::trackFullData $oldName oldInfo
         catalog::trackFullData $newName newInfo
         # oldName is really latest file, newName will be existing file
         set textInfo "
     \tNew ->\tExisting
Len: \t$oldInfo(duration)\t$newInfo(duration)
Rate:\t$oldInfo(Rate)\t$newInfo(Rate)
Size:\t$oldInfo(size)\t$newInfo(size)
"
         if {($oldInfo(size) == $newInfo(size))} {
            set res [tk_messageBox -icon question -type yesno -title "File Exists" \
               -message "Warning: $newName exists\nand appears to be identical, do you still want to overwrite it?"]
         } else {
            set res [tk_messageBox -icon question -type yesno -title "File Exists" \
               -message "Warning: $newName\nexists, do you want to overwrite it?\n$textInfo"]
         }
      }
      if {$res != "yes"} {
         return -code error "$newName File Exists"
      }
   }
   ##
   # Build list of files to copy
   #
   if {[file isdirectory $oldName]} {
      set srcs [fileutils::rglob $oldName]
      set isdir 1
   } else {
      set srcs $oldName
      set isdir 0
   }
   ##
   # Copy the files and add to the database
   # Copy all files found, not just media files
   #
   set errors 0
   set Result ""
   set todo [llength $srcs]
   set counter 0
   foreach {file} $srcs {
      if {$isdir} {

         set target [file join $newName [file tail $file]]
         incr counter
         waitWindow::percent [incr counter] $todo "Copying\n[splitFile $file]" 1
         if {[waitWindow::ifbreak]} {
            saLog "Aborting Copy..."
            Wait 0
            waitWindow::close
            return
         }
      } else {
         set target $newName
      }
      # Attempt to copy file
      if {[catch {bcopy $file $target} result]} {
         append Result "\n$result"
         incr errors
      }  else {
         if {[catalog::isTrackFile $target]} {
            catalog::addNewEntry $target
         }
      }
   }
   if {$errors > 0} {
      return -code error " ($errors) Copy Errors $Result"
   }
}

#------------------------------------------------------------------------------
# Add a new entry to the database and create nodes for it as required
#------------------------------------------------------------------------------
proc catalog::addNewEntry {newName} {
   Trace
   # database if it is a file
   if {[file isfile $newName]} {
      db::cacheFileName $newName 2
      set dir [file dirname $newName]
   } else {
      set dir $newName
   }
   # tefolders & node
   if {[APLExists]} {
      APL::addToRoot $dir 0
   } else {
      aplFolder::setToRoot $dir 0
   }
}
##
# Properties Dialog
#
proc catalog::properties {tracks} {
   set long [list]
   set short [list]
   foreach {track} $tracks {
      if {[file exists $track]} {
         lappend short [file tail $track]
         lappend long $track
      }
   }
   if {[llength $long] <= 0} {
      return
   }
   set i 0
   while {[info command .prop$i] != {}} {incr i}
   set top .prop$i
   toplevel $top
   wm title $top "Track Properties"
   if {[llength $long]==1} {
      set hull [labeledFrame $top.data -text [lindex $short 0]]
      set text [trackTextData [lindex $long 0] Details]
      label $hull.info -text $text -justify left -anchor nw
      pack $hull.info -expand true -fill both
      pack $top.data -expand true -fill both
   } else {
      set tof [tabbedOptionFrame::Create $top.tf -Width 40 -options $short -title "Track Properties"]
      wm protocol $top WM_DELETE_WINDOW "tabbedOptionFrame::close $top.tf;destroy $top"
      pack $tof -expand true -fill both
      set i 0
      foreach {track} $long {
         set text [trackTextData $track Details]
         set hull [tabbedOptionFrame::getHull $tof $i]
         label $hull.info -text $text -justify left -anchor nw
         pack $hull.info -expand true -fill both
         pack $hull -expand true -fill both
         incr i
      }
   }
   update idletasks
   myIcon $top
}
#------------------------------------------------------------------------------
# Function   : catalog::trackTextData
# Description: Return the details of a song in both an array and a text
#              string suitable for balloon help.
# Author     : Tom Wilkason
# Date       : 11/20/2001
#------------------------------------------------------------------------------
proc catalog::trackTextData {fullName _Details} {
   upvar $_Details Details
   set Details(ext) ""
   ##
   # Note: Grab this from a database instead of here if it exists
   #
   set hasTag [db::lookup $fullName Details]
   if {[catch {
      set Details(duration) [fracTime $Details(Duration)]

      set Details(ext) [string trimleft [string toupper [file extension $fullName]] .]
      if {$::snackAmpSettings(autoLevel) && [string is double $Details(sampRate)] && $Details(sampRate)>0 && $Details(Stop)} {
         append endData "Start\t: [format %.3f [expr {$Details(Start)/(1000*$Details(sampRate))}]] sec"
         append endData "\nStop\t: [format %.3f [expr { $Details(Duration)- ($Details(Stop)/(1000*$Details(sampRate))) }]] sec before end\n"
      }
      append endData "Rating\t: "  [mrating $Details(Rating)]
      append endData "\nVotes\t: $Details(Votes)"
      if {$Details(Played) > 0} {
         append endData "\nPlayed\t: "  $Details(Played) " Times"
      } else {
         append endData "\nPlayed\t: Not yet Played"
      }
      if {$Details(LastPlay) > 0} {
         append endData ", Last Played: "  [clock format $Details(LastPlay) -format $::snackAmpSettings(timeString)]
      }
   } result]} {
      return "No Data for $fullName available"
   }
   switch -- $Details(ext) {
      MP3 {append Details(Rate) " kbps"}
      OGG {append Details(Rate) " kbps"}
      WAV {append Details(Rate) " kHz"}
      default {set Details(Rate) "Unknown"}
   }
   append Details(sampRate) " Khz"
   ##
   # File info
   #
   if {[catch {file stat $fullName Stat} result]} {
      return Unknown
   }
   set Details(fSize) "[commify $Stat(size)] Bytes"
   set Details(fDate) [clock format $Stat(mtime) -format $::snackAmpSettings(timeString)]
   ##
   # Balloon help
   #
   set Info         "File\t: [file tail $fullName]\nFolder\t: [file dirname $fullName]\n"
   if {$hasTag} {
      append Info [id3Tag::formatID3Data Details]
      set Details(ID3) $Info
   } else {
      set Details(ID3) {}
   }
   id3Tag::appendIfNotBlank Info "Size\t:"     $Details(fSize)
   id3Tag::appendIfNotBlank Info "Date\t:"     $Details(fDate)
   id3Tag::appendIfNotBlank Info "Duration\t:" $Details(duration)
   id3Tag::appendIfNotBlank Info "Bitrate\t:"  $Details(Rate)
   id3Tag::appendIfNotBlank Info "Freq\t:"     $Details(sampRate)
   if {$::snackAmpSettings(autoLevel)} {
      id3Tag::appendIfNotBlank Info "Gain\t:"  $Details(Gain)
   } else {
      set Details(levelFactor) "Not Set"
   }
   append Info $endData
   return [string trim $Info]
}
#------------------------------------------------------------------------------
# Function   : trackFullData
# Description: Return the details of a song in an array for indexing
#              music files.
# Author     : Tom Wilkason
# Date       : 11/20/2001
#------------------------------------------------------------------------------
proc catalog::trackFullData {fullName _Details} {
   upvar $_Details Details
   ##
   # Retrieve the track sound info
   #
   set Details(size) [file size $fullName]
   soundControl::trackInfo Details $fullName

   if {[catch {gainControl::get $fullName} gain]} {
      set Details(Gain) "--"
   } else {
      set Details(Gain) $gain
   }
   ##
   # Get additional file and ID3 info
   #
   set Details(File)   [file tail $fullName]
   ##
   # ID3 Data, if no tag then set Title = File root name
   #
   set result [id3Tag::id3Label $fullName Details $::snackAmpSettings(ID3ReadOrder)]
   if {$result==0} {
      set Details(Title) [file rootname $Details(File)]
      set Details(Mounted) 1
      set Details(Tag) "--"
   }
   return $result
}
#------------------------------------------------------------------------------
# Function   : DeleteFile
# Description: Delete a file, if it is playing then handle it
# Author     : Tom Wilkason
# Date       : 3/24/2002
#------------------------------------------------------------------------------
proc catalog::DeleteFile {fileName} {
   set currentSound [soundControl::fileName]
   set goNext 0
   if {[string equal $currentSound $fileName]} {
      # This is a trick to see if the song is actually playing
      if {[soundControl::isPlaying]} {
         soundControl::stop
         set goNext 1
      }
      soundControl::reset
   }
   # Now try to delete it
   if {[catch {
      file delete -force $fileName
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -message "Error removing $fileName\n$result"
   } else {
     catalog::removeFile $fileName
   }
   return $goNext
}

#------------------------------------------------------------------------------
# Function   :Remove
# Description: Remove a file from the master and playlists list if it exists
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::removeFile {File} {
   global PLB
   # Remove from the master list
   set index [db::whereInDb $File]
   if {$index >= 0} {
      ##
      # Remove from the master list
      #
      db::removeFile $index
      ##
      # Remove from the Playlist
      #
      LB::removeFile $File
      ##
      # Remove from the APL File List (in case it is there)
      #
      catch {aplLB::RemoveFile $File}
   }
}
#------------------------------------------------------------------------------
# Function   : renameInPlaceDialog
# Description: Bring up a in place editing window on a listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
#------------------------------------------------------------------------------
proc catalog::renameInPlaceDialog {W table {oldName "xxx"}} {
   if {![winfo ismapped $W]} {
      return ""
   }
   ##
   # Need to compute location for dialog
   #
   set active [$W index active]
   set bba [$W bbox $active]
   if {[llength $bba]<4} {
      return ""
   }
   set w  .rndialog
   destroy $w
   toplevel $w -class SnackAmp
   foreach {x y - -} $bba {break}
   # If table, adjust for heading
   if {$table} {
      set y [expr {$y - [winfo y $W]}]
   }
   # Left Hand alignment
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x}]
   set Y [expr {$ry+$y}]
   # Make it look like the text underneath
   set font [$W cget -font]
   set bg [$W cget -background]
   set fg [$W cget -foreground]
   set ::TA(rename) $oldName
   # Set size available width and linespacing+2
   set wid [expr {[string length $oldName]+5}]
   set Wwid [winfo width $W]
   set Wheight [expr {[font metrics $font -linespace]+2}]
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)] -expand true -fill x
   #$w.rn selection range 0 end
   wm withdraw $w
   wm overrideredirect $w 1
   set oldFocus [focus]
   wm transient $w [winfo toplevel $W]
   focus -force $w.rn
   wm geometry $w "${Wwid}x$Wheight+$X+$Y"
   bind $w <Escape> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <FocusOut> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .rndialog]
   wm deiconify $w
   raise $w
   # Force on top if possible
   update idletasks
   ontop::ifpossible $w $w
   tkwait window $w
   ##
   # Back from block
   #
   if {[winfo exists $oldFocus]} {
      focus -force $oldFocus
   }
   # Waiting for window to go away
   return $::TA(rename)
}

#------------------------------------------------------------------------------
# Function   : catalog::RenameEntry
# Description: Replace a file in the master list with another one only if it exists
#              Assume file itself has already been renamed.
# Author     : Tom Wilkason
# BUG: When called in rapid order, keys in tablelist may change from what
# this finds in the database because of renumbering due to sorting.
#------------------------------------------------------------------------------
proc catalog::RenameEntry {oldFile newFile} {
   Trace
   global PLB
   ##
   # If old file doesn't exist, such as incoming data
   # then go ahead and add it to master list if needed.
   # Note: At this point, the actual file is renamed to newFile, so oldFile no longer exists
   # but it is still in the database and other locations
   set index [db::cacheFileName $oldFile 1]
   if {$index < 0} {
      saLog "Could not find $oldFile in database"
      return -1
   }
   set tail [file tail $oldFile]
   ##
   # Replace full name in database, this will also change its index due to auto sorting
   #
   db::renameFile $index $newFile
   ##
   # Update Main Play List
   #
   LB::changePlaylistFiles [list $oldFile $newFile]
   ##
   # Rename any files and their indicies in the active list
   #
   trackList::renameFiles [list $index $oldFile $newFile]
   db::dirtyList
   return 1
}
#----------------------------------------------------------------------------
# Multiple entry rename
#----------------------------------------------------------------------------
proc catalog::RenameEntrys {pairs} {
   Trace
   global PLB
   ##
   # If old file doesn't exist, such as incoming data
   # then go ahead and add it to master list if needed.
   # Note: At this point, the actual file is renamed to newFile, so oldFile no longer exists
   # but it is still in the database and other locations
   set filePairs [list]
   set indexFile [list]
   foreach {oldFile newFile} $pairs {
      set index [db::whereInDb $oldFile]
      if {$index < 0} {
         saLog "Could not find $oldFile in database"
         return -1
      }
      lappend filePairs $oldFile $newFile
      lappend indexFile $index $oldFile $newFile
   }
   if {[llength $filePairs]>0} {
      ##
      # Rename the tracks first
      #
      foreach {oldFile newFile} $filePairs {
         db::renameFile $oldFile $newFile
      }
      ##
      # Update Main Play List
      #
      LB::changePlaylistFiles $filePairs
      ##
      # Rename any files and their indicies in the active list
      #
      trackList::renameFiles $indexFile
   }
   return 1
}

#------------------------------------------------------------------------------
# Function   : catalog::moveFolder
# Description: Move a folder (and its files) to some other place
#              and adjust the APL tree accordingly.
#              folder is existing full name
#              target is new full name
#              only called by APL code
# Author     : Tom Wilkason
# Date       : 3/3/2002
#------------------------------------------------------------------------------
proc catalog::moveFolder {folder target} {
   Trace
   global LB
   # Append / to folder so we don't match
   # Other fol ders with same base name
   # eg. Classic vs. Classic Rock
   set folder [string trimright $folder /]
   set target [string trimright $target /]
   if {[string equal $folder $target]} {
      return
   }
   APL::busyCursor 1
   update idletasks
   if {[catch {catalog::RenameFile $folder $target} result]} {
      debug $result
      APL::busyCursor 0
      return -code error $result
   }
   # Remove old folder name
   APL::delItem $folder
   ##
   # Replace teFolder entries with new name
   #
   set subLen [expr {[string length $folder]-1}]
   set i -1
   set pattern [file join $folder *]
   foreach path [aplFolder::foldersMatchingPattern * 1] {
      incr i
      if {[string match $pattern $path] || [string equal $folder $path]} {
         set newName [string replace $path 0 $subLen $target]
         aplFolder::rename $path $newName
         APL::newItem $newName
      }
   }
   ##
   # Replace each matching entry inList and update playlist and APL file list
   # if affected songs were shown in them.
   #
   set i -1
   # Append to folder for pattern matching
   set i 0
   #W: Function to return indicies of match
   set changed [list]
   foreach {file} [db::filesMatchingPattern $pattern] {
      set newName [string replace $file 0 $subLen $target]
      incr i
      db::renameFile $file $newName
      lappend changed $file $newName
   }
   LB::changePlaylistFiles $changed
   if {$i > 0} {
      db::dirtyList
   }
   APL::busyCursor 0
}

#------------------------------------------------------------------------------
# Copy a single folder
#------------------------------------------------------------------------------
proc catalog::copyFolder {folder target} {
   Trace
   global LB
   # Append / to folder so we don't match
   # Other folders with same base name
   # eg. Classic vs. Classic Rock
   set folder [string trimright $folder /]
   set target [string trimright $target /]
   if {[string equal $folder $target]} {
      return
   }
   APL::busyCursor 1
   update idletasks
   if {[catch {catalog::copyFile $folder $target} result]} {
      debug $result
      APL::busyCursor 0
      return -code error $result
   }
   APL::busyCursor 0
}
#------------------------------------------------------------------------------
# Function   : catalog::moveClipboardFiles
# Description: Move the files in the clipboard to some new folder
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::moveClipboardFiles {target} {
   Trace
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   catch {
      foreach {file} [lflip $tryFiles] {
         if {[lsearch -exact $files $file] < 0 && [file exists $file]} {
            lappend files $file
         }
      }
   }
   set moved 0
   soundControl::latency long
   if {[llength $files] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to move files or folders"
         return 0
      }
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to move:\n[join $files \n]\n to\n$target"]
      if {$result == "yes"} {
         set errors [list]
         ##
         # For each file to be moved, move it
         #
         set todo [llength $files]
         set i 0
         foreach file $files {
            set tail [file tail $file]
            set newName [file join $target $tail]
            ##
            # If source is a folder, move it otherwise
            # move the file
            #
            if {[file isdirectory $file]} {
               if {[catch {
                  if {![string equal $file $target]} {
                     waitWindow::percent [incr i] $todo "Moving\n$file to\n$newName" 1
                     catalog::moveFolder $file $newName
                  }
               } result] \
               } then {
                  debug $result
                  lappend errors $file
               }
            } else {
               if {![string equal $file $newName]} {
                  waitWindow::percent [incr i] $todo "Moving\n$file to\n$newName" 1
                  if {[catch {catalog::RenameFile $file $newName} result]} then {
                     debug $result
                     lappend errors $file
                  } else {
                     #W: This can be slow on big moves, you have a full lookup
                     # for each file moved.
                     catalog::RenameEntry $file $newName
                     incr moved
                  }
               }
            }
            if {[waitWindow::ifbreak]} {
               break
            }
         }
         waitWindow::close
         # Handle any errors
         if {[llength $errors] > 0} {
            tk_messageBox -type ok -icon warning -message "The following files were not moved\n[join $errors \n]"
         }
      }
   }
   if {$moved > 0} {
      db::dirtyList
   }
   soundControl::latency normal
   return $moved
}
#----------------------------------------------------------------------------
# Copy clipboard files to a target
#----------------------------------------------------------------------------
proc catalog::copyClipboardFiles {target} {
   Trace
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   catch {
      foreach {file} [lflip $tryFiles] {
         if {[lsearch -exact $files $file] < 0 && [file exists $file]} {
            lappend files $file
         }
      }
   }
   set moved 0
   soundControl::latency long
   if {[llength $files] > 0} {
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to copy:\n[join $files \n]\n to\n$target"]
      if {$result == "yes"} {
         set errors [list]
         ##
         # For each file to be moved, move it
         #
         set todo [llength $files]
         set i 0
         foreach file $files {
            set tail [file tail $file]
            set newName [file join $target $tail]
            waitWindow::percent [incr i] $todo "Copying\n$tail" 1
            ##
            # If source is a folder, move it otherwise
            # move the file
            #
            if {[file isdirectory $file]} {
               if {[catch {
                  if {![string equal $file $target]} {
                     catalog::copyFolder $file $newName
                  }
               } result] \
               } then {
                  debug $result
                  lappend errors $file
               }
            } else {
               if {![string equal $file $newName]} {
                  if {[catch {catalog::copyFile $file $newName} result]} then {
                     debug $result
                     lappend errors $file
                  } else {
                     incr moved
                  }
               }
            }
            if {[waitWindow::ifbreak]} {
               saLog "Copy Aborted..."
               break
            }
         }
         waitWindow::close
         # Handle any errors
         if {[llength $errors] > 0} {
            tk_messageBox -type ok -icon warning -message "The following files were not copied\n[join $errors \n]"
         }
      }
   }
   if {$moved > 0} {
      db::dirtyList
   }
   soundControl::latency normal
   return $moved
}
##
# Return 1 if the folder is contained in a media player
#
proc catalog::isFolderMediaPlayer {dest} {
   foreach {folder} $::snackAmpSettings(mp3Player) {
      if {[string match $folder* $dest]} {
         return 1
      }
   }
   return 0
}
#----------------------------------------------------------------------------
# Return 1 if the folder is one of the fixed ones
#----------------------------------------------------------------------------
proc catalog::isFixedFolder {dest} {
   foreach {folder} $::snackAmpSettings(mediaDirs) {
      if {[string match $folder* $dest]} {
         return 1
      }
   }
   return 0
}
#----------------------------------------------------------------------------
# Return 1 if the folder is one of the fixed ones
#----------------------------------------------------------------------------
proc catalog::isMountedFolder {dest} {
   foreach {folder} $::TA(mediaDirs) {
      if {[string match $folder* $dest]} {
         return 1
      }
   }
   return 0
}
#----------------------------------------------------------------------------
# Return the icon type for a path, the list numbers are assigned in catalog (HACK)
#----------------------------------------------------------------------------
proc catalog::getIconType {folder} {
   global TA
   foreach {Path type icon} $TA(mediaDirPair) {
      if {[string match ${Path}* $folder]} {
         return $icon
      }
   }
   return 0
}
##
# Return 1 if the destination is a media player and the source is not
#
proc catalog::isMediaPlayer {dest {source ""}} {
   Trace
   set idest 0
   foreach {folder} $::snackAmpSettings(mp3Player) {
      if {[string match [file join $folder *] $dest] || [string equal $folder $dest]} {
         set idest 1
      } else {
         set idest 0
      }
      if {[string length $source] && [string length $folder]} {
         set isource [string match [file join $folder *] $source]
         if {$idest==1 && $isource==0} {
            return 1
         } else {
            set $idest 0
         }
      } elseif {$idest} {
         return 1
      }
   }
   return $idest
}
#------------------------------------------------------------------------------
# Function   : renameFolder
# Description: Prompt then rename a subfolder under some level
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::renameFolder {folder} {
   set tail [file tail $folder]
   set base [file dirname $folder]
   set newName $tail
   set res 0
   if {[getEntryDialog $newName newName "Rename Folder"]=="ok" && ![string equal $tail $newName]} {
      if {[catch {
         set newFullName [file join $base $newName]
         moveFolder $folder $newFullName
         if {[APLExists]} {
            APL::treeSelect $newFullName
            APL::viewHighlighted $::TREE
         }
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder $newName\n$result"
         set res 0
      }
   }
   return $res
}

#------------------------------------------------------------------------------
# Function   : removeFolder
# Description: Remove a folder and all of its children files
#              Only used by APL tree
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::removeFolder {folder} {
   set i 0
   if {$::snackAmpSettings(dontMoveFiles)} {
      tk_messageBox -icon warning -type ok \
         -message "You don't have permission to remove folders"
      return
   }

   set pattern [file join $folder *]
   # If a removeable file in name then check if at base level and if so
   # verify before removing the catalog file from disk
   set contents [glob -nocomplain $pattern]
   set i [llength $contents]  ;# don't just check music tracks
   set parts [file split $folder]
   set message "Are you sure you want to remove $folder and its decendents, it contains $i files\n\nAll files and folders will be permanently deleted!"
   if {$i == 0 || [tk_messageBox -icon warning -title "Please Respond" -type yesno -message $message] == "yes"} {
      ##
      # Handle the files
      #
      foreach {file} $contents {
         if {[catch {
            DeleteFile $file
         } result]} then {
            debug $result
         }
      }
      ##
      # Handle the folders, work from the longest folder back
      # to maximize the deleted folders
      foreach path [lsort -decreasing [aplFolder::foldersMatchingPattern $folder*]] {
         if {[string match $pattern $path] || [string equal $folder $path]} {
            if {[catch {
               file delete -force $path
               aplFolder::Unset $path
               APL::delItem $path
            } result]} then {
               debug $result
               return
            }
         }
      }
      APL::delItem $folder
      APL::treeSelect [file dirname $folder]
      aplLB::UpdateList [file dirname $folder]
      gui::updateFileDisplay
   }
   return
}
#------------------------------------------------------------------------------
# Function   : catalog::SavePlaylistCache
# Description: Cache the current playlist for later restoration
# Author     : Tom Wilkason
# TBD: Save/restore extended info if it is faster than a lookup
#------------------------------------------------------------------------------
proc catalog::SavePlaylistCache {args} {
   global TA
   set sList [LB::keyAndNames]  ;# key|file pairs
   set loc [LB::idxPlaying]
   if {$loc < 0} {
      set loc [LB::idxActive]
   }
   append sList "\n$loc"
   append sList "\n[soundControl::samplePosition]"
   append sList "\n$soundControl::playState"
   if {[catch {safeWrite $TA(playCacheFile) $sList} result]} {
      return -code error "Save Error: Unable to open '$TA(playCacheFile)' for writing\n"
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::ReadPlaylistCache
# Description: Restore the playlist from a cached file (during initialization)
# Author     : Tom Wilkason
#------------------------------------------------------------------------------
proc catalog::ReadPlaylistCache {args} {
   global TA
   ##
   # If new version (format) exists, use it, otherwise try old version
   #
   if {[file exists $TA(playCacheFile)]} {
      set pairs 1
      set playCacheFile $TA(playCacheFile)
   } else {
      set playCacheFile [string trimright $TA(playCacheFile) "3"] ;# try old style
      set pairs 0
   }
   if {[catch {open $playCacheFile r} InFile]} {
      return -code error "Read Error: Unable to open '$playCacheFile' for reading\n"
   }
   fconfigure $InFile -buffersize 300000
   set sList  [gets $InFile]
   set cIndex [gets $InFile]
   # Restore position within the file
   if {[eof $InFile]} {
      set cPosition -1
   } else {
      set cPosition [gets $InFile]
   }
   if {[eof $InFile]} {
      set playState "Stopped"
   } else {
      set playState [gets $InFile]
   }
   close $InFile
   # Replace current playlist very fast
   LB::deleteAll
   if {[llength $sList] > 0} {
      if {$pairs} {
         LB::replacePairList $sList
      } else {
         LB::insertList end $sList
      }
      if {$::snackAmpSettings(resumePlay)} {
         if {$playState=="Playing"} {
            after idle catalog::resumePlay $cIndex $cPosition
         } elseif {$playState=="Paused"} {
            after idle catalog::resumePlay $cIndex $cPosition 1
         } else {
            macros::Highlight $cIndex
         }
      } else {
         macros::Highlight $cIndex
      }
      saLog "Restored playlist cache '$pairs' '$playCacheFile' of [llength $sList] files"
   }
}
##
# Function to resume play
#
proc catalog::resumePlay {index offset {pause 0}} {
   macros::Highlight $index
   Play active $offset $pause
}
#------------------------------------------------------------------------------
# Return a list of all AutoPlaylist files
#------------------------------------------------------------------------------
proc catalog::getAutoPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.tafs]]
}
#------------------------------------------------------------------------------
# Return a list of all playlist files
#------------------------------------------------------------------------------
proc catalog::getPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.taf *.m3u]]
}
#------------------------------------------------------------------------------
# Return a "yes" if using folder names, "no" if tag info
# ask user if not known.
#------------------------------------------------------------------------------
proc catalog::getMediaMode {mp} {
   global snackAmpSettings
   if {![info exists snackAmpSettings(mediaMode,$mp)]} {
      set result [tk_messageBox -type yesno -title "Please respond" -icon question \
         -message "Please answer yes to use folder structure to copy, \nno to use ID3 Artist/Album for $mp"]
      set snackAmpSettings(mediaMode,$mp) $result
   }
   return $snackAmpSettings(mediaMode,$mp)
}
#----------------------------------------------------------------------------
# Utility routines
#----------------------------------------------------------------------------
proc tester {{num 5}} {
   set TRACKS [db::sortedList]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file dirname $file]
      }
   } $num]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file root $file]
      }
   } $num]
   puts [time {
      foreach {file} $TRACKS {
         set dir [file tail $file]
      }
   } $num]
}
#------------------------------------------------------------------------------
# Function   : catalog::ShowDups
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::ShowDups {{pattern *}} {
   foreach {file} [db::filesMatchingPattern $pattern] {
      set dir [file dirname $file]
      set tail "$dir/[lrange [split $file -] end-1 end]"
      set fileTail [lrange [file split $file] end-1 end]
      if {![info exists tailHash($tail)]} {
         set tailHash($tail) $file
      } else {
         update
         saLog "Duplicate $tail\n   '$file'\n   '$tailHash($tail)'\n"
      }
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::DupFiles
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::DupFiles {{pattern *}} {
   foreach {file} [db::sortedList] {
      if {[string match $pattern $file] && ![string match $patt $file]} {
         set tail [lindex [file split $file] end]
         if {![info exists tailHash($tail)]} {
            set tailHash($tail) $file
         } else {
            update
            saLog "Duplicate $tail\n   '$file - [file size $file]'\n   '$tailHash($tail) - [file size $tailHash($tail)]'\n"
         }
      }
   }
}
#------------------------------------------------------------------------------
# Function   : catalog::DupSizes
# Description: Command line routine to show potential duplicate files
#              based on like file sizes. Good for finding renamed files.
# Author     : Tom Wilkason
# Date       : 3/28/2002
#------------------------------------------------------------------------------
proc catalog::DupSizes {{pattern *} {threshold 2}} {
   foreach {file} [db::filesMatchingPattern $pattern] {
      if {![catch {set tail [file size $file]} result]} then {
         lappend tailHash($tail) $file
      }
   }

   foreach entry [array names tailHash] {
      set files  $tailHash($entry)
      if {[llength $files] > 1} {
         set last [expr {[llength $files]-1}]
         set matches 0
         # Pass through each set of duplicate files
         for {set i 0} {$i < $last} {incr i} {
            set base [split [string map {" " ""} [string tolower [file tail [lindex $files $i]]]] -]
            set next [split [string map {" " ""} [string tolower [file tail [lindex $files [incr i]]]]] -]
            foreach part $base {
               if {[lsearch -exact $next $part] > -1} {
                  incr matches
               }
            }
         }
         if {$matches > $threshold} {
            puts "Duplicates found for size $entry ($matches)"
            foreach file $files {
               puts "   $file"
            }
         }
      }
   }
}
#------------------------------------------------------------------------------
# Read the cached files into memory
# Set up the default teFolders
# Returns a list of paths not found
#------------------------------------------------------------------------------
proc catalog::convertCacheFiles {args} {
   global TA snackAmpSettings
   ##
   # Restore the cached files from each media Dir
   # Support either new type (a list) or old type (Array)
   set result [list]
   # Move list into a hash table for faster access
   # and to allow replacing parts by single path
   #catalog::validateMediaPaths
   foreach {Path} [lsort $TA(mediaDirs)] {
      waitWindow::show "Importing Directory $Path - Please Wait"
      # Read either v1 or v2 cache files
      if {![file exists $Path]} {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "Warning" \
            -message "Media Folder '$Path' does not exist,\nit may be named improperly or on \nan unavailable network computer."

      } else {
         if {[catch {set fid [open [file join $Path $TA(cacheFile)] r]}]} {
            lappend result $Path
         } else {
            # V2 files
            fconfigure $fid -buffersize 100000
            gets $fid tLlist
            close $fid
            saLog "Converting $Path (v2)"
            # add the new ones for this path back
            set i 0
            set total [llength $tLlist]
            foreach {File} $tLlist {
               set file [file join $Path $File]
               incr i
               if {$i%25==0} {
                  waitWindow::percent $i $total "Importing\n$Path" 1
                  if {[waitWindow::ifbreak]} {
                     db::commit
                     return
                  }
               }
               db::cacheFileName $file 1
            }
            db::commit
         }
      }
   }
   return
}
#------------------------------------------------------------------------------
# Function   : fixplus::correctCase
# Description: Replace + in file names (due to http download), also fix other
#              case problems and clean up spacing.
# Author     : Tom Wilkason
# Date       : 1/8/2001
#------------------------------------------------------------------------------
namespace eval fixplus  {}
proc fixplus::correctCase {folder} {
   set changed 0
   # Use a folder or the selected names
   if {[file isdirectory $folder]} {
      set Files [glob -nocomplain -directory $folder -types {f l} *]
   } else {
      set Files $folder
   }
   foreach File $Files {
      set folder [file dirname $File]
      set File   [file tail $File]
      ##
      # Clean up the file name
      #
      set ext [file extension $File]
      set root [file rootname $File]   ;# base part of name
      if {[catalog::isTrackFile $File]} {
         # Clean out funny stuff
         set newRoot [cleanName $root]
         set ofn [file join $folder $File]
         set fnf [file join $folder "$newRoot$ext"] ;#full new file
         ##
         # Have new file name
         #
         if {![string equal $newRoot $root]} {
            ##
            # Try to rename the file here
            #
            if {[catch {catalog::RenameFile $ofn $fnf} result]} {
               saLog "correctCase Error:$result"
            } else {
               catalog::RenameEntry $ofn $fnf
               incr changed
            }
         }
      }
   }
   if {$changed > 0} {
      db::dirtyList
   }
   return $changed
}
##
# Make sure the file name as valid characters
#
proc fixplus::cleanFileName {root} {
   set root [string map {[ ( ] ) \{ ( \} ) } $root]
   return $root
}
#------------------------------------------------------------------------------
# Function   : cleanName
# Description: Clean up a file name to use - as separator
# Author     : Tom Wilkason
# Date       : 3/19/2002
#------------------------------------------------------------------------------
proc fixplus::cleanName {root} {
   # sp is user supplied split char, "-" is recommended
   set sp [string trim $::snackAmpSettings(splitChar)]
   set root [string map {[ ( ] )} $root]
   regsub -all -- {[+_ ]+} $root { } root
   regsub -all -- {( *- *)|( *\+ *)} $root " $sp " root
   set root [string trim $root]
   # Fix first part if a number
   set reExp [subst -nocommand {^([0-9]+)[$sp ]*(.*?)}]
   if {[regexp -- $reExp $root -> num rest]} {
      set root "$num $sp $rest"
   }
   # Cap each token
   set root [stringTitle $root]
   return $root
}
#  ----------------------------------------------------------------------------
#  aplFolder namespace: Wrap access to the AutoPlaylist folder tree settings
#  ----------------------------------------------------------------------------

namespace eval aplFolder {
}
#------------------------------------------------------------------------------
# Function   : getAPLFileList
# Description: Return a list of all APL files
# Author     : Tom Wilkason
# Date       : 12/24/2002
#------------------------------------------------------------------------------
proc aplFolder::getAPLFileList {args} {
   set result [list]
   # TODO: This would be faster to lsort -unique a list of folders
   # or to flag all files and query them
   foreach {File} [db::sortedList] {
      if {[isSet [file dirname $File]]} {
         lappend result $File
      }
   }
   return $result
}

#------------------------------------------------------------------------------
# Function   : foldersMatchingPattern
# Description: Return a list of folders in or under a path
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::foldersMatchingPattern {{pattern *} {sorted 0}} {
   if {$sorted} {
      return [lsort [array names ::teFolders $pattern]]
   } else {
      return [array names ::teFolders $pattern]
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::getFolders
# Description: Return a list of some folders and subfolders of subfolders 'subfolders'
# Author     : Tom Wilkason
# Date       : 2/16/2003
#------------------------------------------------------------------------------
proc aplFolder::getFolders {selectedFolder {subfolders 0}} {
   lappend folders $selectedFolder
   # Find any remaining subfolders
   if {$subfolders>0} {
      set baseDepth [llength [file split $selectedFolder]]  ;#0 is just this folder
      foreach {index} [aplFolder::foldersMatchingPattern $selectedFolder/*] {
         set thisDepth [llength [file split $index]]
         if {$thisDepth <= ($baseDepth+$subfolders)} {
            lappend folders $index
         }
      }
   }
   # It would be nice to have -dictionary, but the sorted search will fail
   return [lsort $folders]
}
#------------------------------------------------------------------------------
# Function   : folderArray
# Description: Return an array (index element) of folders in or under a path
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::folderArray {{pattern *}} {
   return [array get ::teFolders $pattern]
}
#------------------------------------------------------------------------------
# Function   : isSet
# Description: Return true if the folder item is set (and exists)
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::isSet {dirName {value 1}} {
   if {[info exists ::teFolders($dirName)] && ($::teFolders($dirName) == $value)} {
      return 1
   } else {
      return 0
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Exists
# Description: Veriy the entry or the actual folder exists
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Exists {Path {check 0}} {
   if {$check} {
      return [expr {[info exists ::teFolders($Path)] || [file isdirectory $Path]}]
   } else {
      return [info exists ::teFolders($Path)]
   }
}
#------------------------------------------------------------------------------
# Function   : Unset
# Description: Remove one or moer entries
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Unset {key} {
   unset -nocomplain ::teFolders($key)
}

#------------------------------------------------------------------------------
# Verify and remove all nodes that no longer exist
#------------------------------------------------------------------------------
proc aplFolder::removeUnusedNodes {pattern} {
   global teFolders
   foreach {key} [array names teFolders $pattern] {
      if {![file exists $key]} {
         unset teFolders($key)
         if {[APLExists]} {
            APL::delItem $key
         }
      }
      update
   }
}
#------------------------------------------------------------------------------
# Function   : stripBranch
# Description: Remove one or moer entries
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::stripBranch {pattern} {
   array unset ::teFolders $pattern
}
#------------------------------------------------------------------------------
# Function   : SetSome
# Description: Set some values to a value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetSome {pattern value} {
   foreach {key} [array names ::teFolders $pattern] {
      set ::teFolders($key) $value
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Set
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Set {key value} {
   set ::teFolders($key) $value
}
#------------------------------------------------------------------------------
# Function   : aplFolder::SetIfBlank
# Description: Set an entry to some value only if it doesn't already exist
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetIfBlank {key value} {
   if {![info exists ::teFolders($key)]} {
      set ::teFolders($key) $value
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::SetIfExists
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::SetIfExists {key value} {
   if {[info exists ::teFolders($key)]} {
      set ::teFolders($key) $value
   }
}
#------------------------------------------------------------------------------
# Function   : aplFolder::Get
# Description: Set an entry to some value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::Get {key} {
   return $::teFolders($key)
}
#------------------------------------------------------------------------------
# Function   : aplFolder::rename
# Description: Rename an entry retaining its old value
# Author     : Tom Wilkason
# Date       : 8/6/2003
#------------------------------------------------------------------------------
proc aplFolder::rename {path newName} {
   set value $::teFolders($path)
   unset ::teFolders($path)
   set ::teFolders($newName) $value
}
#----------------------------------------------------------------------------
# Return the number of tree entries
#----------------------------------------------------------------------------
proc aplFolder::entries {args} {
   return [array size ::teFolders]
}
proc aplFolder::savePattern {pattern} {
   return [array get ::teFolders $pattern]
}
proc aplFolder::restoreNames {saveNames} {
   foreach {folder value} $saveNames {
      aplFolder::SetIfExists $folder $value
   }
}
#------------------------------------------------------------------------------
# Function   : db::createTeFolders
# Description: Create all the teEntries and set them to 0 for the
#              database.
#              if a parent doesn't exist, create one and set it to 0
# Author     : Tom Wilkason
# Date       : 8/8/2003
#
#------------------------------------------------------------------------------
# bug this is 10x slower when using metakit, looks like file dirname is much slower
# not much faster than using db::sortedList
proc aplFolder::createTeFolders {{how 0}} {
   Trace
   global teFolders
   set mediaPatterns [list]
   foreach {dir} $::TA(mediaDirs) {
      lappend mediaPatterns [file join $dir *]
   }
   set hits 0
   set lastFolder {}
   # Looping is very fast, the [file dirname] is 3x slower
   # move to database
   $db::dbObj loop row {
      foreach {folder Mounted} [$db::dbObj get $row Folder Mounted] {break}
      ##
      # Build the folder tree entries
      #
      if {$Mounted==1 && [string equal $folder $lastFolder]==0} {
         incr hits [aplFolder::setToRoot $folder $how]
      } else {
         set lastFolder $folder
      }
   }
   return $hits
}
##
# Check to see if a track is in a valid media folder
#
proc aplFolder::isInMediaDir {file mediaPatterns} {
   foreach {patt} $mediaPatterns {
      if {[string match $patt $file]} {
         return 1
      }
   }
   return 0
}
##
# Ensure we have a path to the root
#
proc aplFolder::setToRoot {folder how} {
   global teFolders
   set hits 0
   while {![info exists teFolders($folder)]} {
      incr hits
      set teFolders($folder) $how
      # Backup one
      set root [file dirname $folder]
      if {[string equal $folder $root]} {
         break
      } else {
         set folder $root
      }
   }
   return $hits
}
#------------------------------------------------------------------------------
# Function   : setTeDefault (recursive)
# Description: Fixup a branch of teFolders that may be undetermined
#              Work backwards and set the childrens value to that
#              of the parent all the way to the root. Default to 0
# Author     : Tom Wilkason
# Date       : 3/24/2001
#------------------------------------------------------------------------------
proc aplFolder::setTeDefault {Index} {
   global teFolders
   switch -- $teFolders($Index) {
      -1 {
         # Folder wasn't in new APL file,
         # Set it to it's parents value if possible
         set parent [file dirname $Index]
         if {$parent != $Index} {
            if {[info exists teFolders($parent)]} {
               set teFolders($Index) [aplFolder::setTeDefault $parent]
            } else {
              set teFolders($Index) 0
            }
         } else {
            set teFolders($Index) 0
         }
      }
      1 {
         # Folder is set on, if the parent was set off change
         # the parent to a partial state (we know it can't be off)
         set parent [file dirname $Index]
         if {($parent != $Index) && [aplFolder::isSet $parent 0]} {
            set teFolders($parent) 2
         }
      }
      2 {
         # If it was set partial before, just turn it off
         set teFolders($Index) 0
      }
      default {}
   }
   return $teFolders($Index)
}
##
# Verify each tefolder entry (very slow)
#
proc aplFolder::check {args} {
   foreach {folder} [array names ::teFolders] {
      if {![file isdirectory $folder]} {
         unset ::teFolders($folder)
      }
   }
}
##
# Save/restore the folder cache
#
proc aplFolder::saveFolderNames {args} {
   safeWrite $::TA(folderCacheFile) [array get ::teFolders]
}
proc aplFolder::restoreFolderNames {args} {
   global TA
   if {[file exists $TA(folderCacheFile)]} {
      if {[catch {set fid [open $TA(folderCacheFile) r]} result]} {
         saLog $result
         return 0
      } else {
         if {[catch {
            fconfigure $fid -buffersize 300000
            set data [gets $fid]
            close $fid
            array set ::teFolders $data
         } result] } then {
            saLog $result
            return 0
         }
      }
   }
   return 1
}
proc aplFolder::showWithoutCoverArt {pattern} {
   foreach {folder} [foldersMatchingPattern $pattern 1] {
      update
      if {[llength [foldersMatchingPattern [file join $folder *]]]==0} {
         foreach {image valid} [coverart::covername $folder] {break}
         if {![string length $image]} {
            puts "$folder"
         }
      }
   }
}

##
# Create an html page and launch the browser on it for the
# files on the clipboard
#
proc html {args} {
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   # Ignore clibboard entries that aren't files
   # Ignore clibboard entries are duplicates
   # Ignore errors due to funky clipboard data
   set html ""
   catch {
      foreach {file} [lflip $tryFiles] {
         if {[db::lookup $file Data 1]>=3} {
            append html "<TR><TD>$Data(Artist)</TD>$Data(Title)<TD>[timify $Data(Duration)]</TD></TR>\n"
         }
      }
   }
   clipboard append $html
   return 1
}

package provide catalog 1.0
