#------------------------------------------------------------------------------
# remoteif Package 1.0
#
# client interface for a xmms like remote interface for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2005 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
# Estable a port server to accept commands via a socket
# TODO: Add code to require USER/PASS pair to enable the socket state
#----------------------------------------------------------------------------
namespace eval remoteif {
   variable socket 0
   variable port
   variable host
   variable useAPI
   upvar #0 snackAmpSettings(useAPI)     useAPI
   upvar #0 snackAmpSettings(ircHost)    host
   upvar #0 snackAmpSettings(apiPort)    port
}
#----------------------------------------------------------------------------
# Install the remote control socket interface
#----------------------------------------------------------------------------
proc remoteif::Server {args} {
   variable port
   variable host
   variable useAPI
   if {$useAPI} {
      if {[string length [string trim $host]]==0} {
         set host [info hostname]
      }
      if [catch {socket -server remoteif::Accept $port} result] {
         saLog "Can't establish API server for $host on port $port\n$result"
      } else {
         saLog "API Server Established on port $port"
      }
   }
}
#----------------------------------------------------------------------------
# Setup a socket and callback for this connection
#----------------------------------------------------------------------------
proc remoteif::Accept {sock addr port} {
   fconfigure $sock -buffering line -translation auto
   fileevent $sock readable [list remoteif::Process $sock]
}
#########################################################################
# Function    : remoteif::Process
# Description : Called from fileevent on a readable input on my port
# Author      : Tom Wilkason
# Date        : 6/4/2005
#########################################################################
proc remoteif::Process {sock} {
   global snackAmpSettings
   variable port
   variable host
   ;# If eof or partial line the close up the channel now
   set result "EOF"
   if {[eof $sock] || [catch {gets $sock line} result]} {
      # end of transmission or abnormal connection drop
#puts "$sock closed"
      # clean up session info for this socket
      catch {close $sock}
   } else {
      # xmms_remote subcommand {param list}
      # subcommand params
#saLog $line
      # ensure this socket has been "logged in" before accepting

      set subcommand [lindex $line 0]
      set params [lrange $line 1 end]
      puts $sock [process_xmms_remote $subcommand $params]
   }
}
#------------------------------------------------------------------------------
# Process two way xmms_remote functions, stay close to xmms remote control
# protocol
#------------------------------------------------------------------------------
proc process_xmms_remote {subcommand paramlist} {
   saLog "$subcommand $paramlist"
   set error "OK"
   set cc ""
   switch -- $subcommand {
      xmms_remote_get_info {
         set info [soundControl::sinfo]
         set rate [soundControl::bitrate]
         set freq [lindex $info 1]
         set chan [lindex $info 5]
         set cc "$rate $freq $chan"
      }
      xmms_remote_get_output_time {
         set cc [expr {round([soundControl::position]*1000)}]
      }
      xmms_remote_get_playlist_pos {
         set cc [LB::idxPlaying]
      }
      xmms_remote_get_playlist_file {
         set cc [LB::currentTrack]
      }
      xmms_remote_get_playlist_length {
         set cc [LB::idxEnd]
      }
      xmms_remote_get_playlist_title {
         set pos [lindex $paramlist 0]
         if {[string is integer -strict $pos]} {
            set track [LB::keyToName [LB::keyFromIndex $pos]]
         } else {
            set track [LB::currentTrack]
         }
         db::lookup $track Data 1
         set cc Data(Title)
      }
      xmms_remote_is_running {
         set cc 1
      }
      xmms_remote_set_playlist_pos {
         set pos [lindex $paramlist 0]
         set cc [LB::activate $pos]]
      }
      xmms_remote_play {
         Play
      }
      xmms_remote_playlist_add {
         # append a list of tracks
         set files $paramlist
         # TODO: need to parse this list of files
         LB::AppendList $paramlist
      }
      xmms_remote_set_main_volume {
         set amount [lindex $paramlist 0]
         macros::PlayGain [expr {$amount/60.0*100.0}] ;# max is 60
      }
      xmms_remote_get_main_volume {
         set cc [macros::getPlayGain]
      }
      xmms_remote_stop {
         Stop
      }
      xmms_remote_pause {
         Pause
      }
      xmms_remote_playlist_next {
         Next
      }
      xmms_remote_playlist_prev {
         Prev
      }
      xmms_remote_jump_to_time {
         set newTime [lindex $paramlist 0]
         gui::setTime $newTime
         PlayOffset
      }
      xmms_remote_get_playlist_time {
         set cc [soundControl::duration]
      }
      xmms_remote_is_playing {
         set cc [expr {$soundControl::playState=="Playing"}]
      }
      xmms_remote_is_paused {
         set cc [expr {$soundControl::playState=="Paused"}]
      }
      xmms_remote_quit {
         # let's not for now
      }
      default {
         set cc "Unknown command: $subcommand"
         set error "ERROR"
      }
   }
   return  "$error $cc"
}
#----------------------------------------------------------------------------
# Fire up the API server
#----------------------------------------------------------------------------
remoteif::Server
package provide remoteif 1.0


